--[[
Copyright (C) GtX (Andy), 2022

Author: GtX | Andy
Date: 18.03.2022
Revision: FS22-01

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

About:
Spec created for Johnny Vee to maintain a 500 cow farm without unrealistic Lely Vector fill levels
]]

FeedingRobotInterval = {}

FeedingRobotInterval.MOD_NAME = g_currentModName
FeedingRobotInterval.SPEC_NAME = string.format("spec_%s.feedingRobotInterval", g_currentModName)

function FeedingRobotInterval.registerXMLPaths(schema, basePath)
    schema:setXMLSpecializationType("Husbandry")
    schema:register(XMLValueType.INT, "placeable.husbandry.feedingRobot#feedsPerHour", "Number of feeds per hour. Possible options: 2, 3, 4, 5, 6")
    schema:setXMLSpecializationType()
end

function FeedingRobotInterval.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(PlaceableHusbandryFeedingRobot, specializations)
end

function FeedingRobotInterval.registerEventListeners(placeableType)
    SpecializationUtil.registerEventListener(placeableType, "onPostLoad", FeedingRobotInterval)
    SpecializationUtil.registerEventListener(placeableType, "onMinuteChanged", FeedingRobotInterval)
end

function FeedingRobotInterval.registerOverwrittenFunctions(placeableType)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "getNeedMinuteChanged", FeedingRobotInterval.getNeedMinuteChanged)
end

function FeedingRobotInterval:onPostLoad(savegame)
    self.spec_feedingRobotInterval = self[FeedingRobotInterval.SPEC_NAME]

    if self.spec_feedingRobotInterval == nil then
        g_logManager:error("[%s] Specialization with name 'feedingRobotInterval' was not found in modDesc!", FeedingRobotInterval.MOD_NAME)

        SpecializationUtil.removeEventListener(self, "onMinuteChanged", FeedingRobotInterval)
        return
    end

    local spec = self.spec_feedingRobotInterval

    spec.feedsPerHour = 1
    spec.intervalMinutes = 60

    if self.spec_husbandryFeedingRobot.feedingRobot ~= nil then
        -- Possible intervals: 60min, 30min, 20min, 15min, 12min, 10min..
        spec.feedsPerHour = MathUtil.clamp(self.xmlFile:getValue("placeable.husbandry.feedingRobot#feedsPerHour", 3), 2, 6)

        spec.intervalMinutes = 60 / spec.feedsPerHour
    else
        g_logManager:error("[%s] No 'feedingRobot' found!", FeedingRobotInterval.MOD_NAME)
    end

    if spec.intervalMinutes == 60 then
        SpecializationUtil.removeEventListener(self, "onMinuteChanged", FeedingRobotInterval)
    end
end

function FeedingRobotInterval:onMinuteChanged(minute)
    if self.isServer and minute ~= 0 then
        if minute % self.spec_feedingRobotInterval.intervalMinutes == 0 then
            local feedingRobot = self.spec_husbandryFeedingRobot.feedingRobot
            local robot = feedingRobot.robot

            if robot ~= nil and robot.recipe ~= nil and self.spec_husbandryFood:getFreeFoodCapacity(robot.recipe.fillType) > robot.fillPlane.capacity then
                feedingRobot:createFoodMixture()
            end
        end
    end
end

function FeedingRobotInterval:getNeedMinuteChanged(superFunc)
    if self.spec_husbandryFeedingRobot ~= nil and self.spec_husbandryFeedingRobot.feedingRobot ~= nil then
        return true
    end

    return false
end
