PlaceableHusbandryGoatMilk = {}

source("dataS/scripts/animals/husbandry/objects/MilkingRobot.lua")

function PlaceableHusbandryGoatMilk.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(PlaceableHusbandry, specializations)
end

PlaceableHusbandryGoatMilk.currentModName = g_currentModName;

function PlaceableHusbandryGoatMilk.registerFunctions(placeableType)
	SpecializationUtil.registerFunction(placeableType, "onMilkingRobotLoaded", PlaceableHusbandryGoatMilk.onMilkingRobotLoaded)
end

function PlaceableHusbandryGoatMilk.registerOverwrittenFunctions(placeableType)
	SpecializationUtil.registerOverwrittenFunction(placeableType, "updateOutput", PlaceableHusbandryGoatMilk.updateOutput)
	SpecializationUtil.registerOverwrittenFunction(placeableType, "updateProduction", PlaceableHusbandryGoatMilk.updateProduction)
	SpecializationUtil.registerOverwrittenFunction(placeableType, "updateInfo", PlaceableHusbandryGoatMilk.updateInfo)
	SpecializationUtil.registerOverwrittenFunction(placeableType, "getConditionInfos", PlaceableHusbandryGoatMilk.getConditionInfos)
end

function PlaceableHusbandryGoatMilk.registerEventListeners(placeableType)
	SpecializationUtil.registerEventListener(placeableType, "onLoad", PlaceableHusbandryGoatMilk)
	SpecializationUtil.registerEventListener(placeableType, "onDelete", PlaceableHusbandryGoatMilk)
	SpecializationUtil.registerEventListener(placeableType, "onFinalizePlacement", PlaceableHusbandryGoatMilk)
	SpecializationUtil.registerEventListener(placeableType, "onHusbandryAnimalsUpdate", PlaceableHusbandryGoatMilk)
end

function PlaceableHusbandryGoatMilk.registerXMLPaths(schema, basePath)
	schema:setXMLSpecializationType("Husbandry")

	basePath = basePath .. ".husbandry.goatmilk"

	schema:register(XMLValueType.NODE_INDEX, basePath .. ".milkingRobots.milkingRobot(?)#linkNode", "Milkingrobot link node")
	schema:register(XMLValueType.STRING, basePath .. ".milkingRobots.milkingRobot(?)#class", "Milkingrobot class name")
	schema:register(XMLValueType.STRING, basePath .. ".milkingRobots.milkingRobot(?)#filename", "Milkingrobot config file")
	schema:setXMLSpecializationType()
end

function PlaceableHusbandryGoatMilk:onLoad(savegame)
	self.spec_husbandryGoatMilk = {}; 
	local spec = self.spec_husbandryGoatMilk
	spec.litersPerHour = 0
	spec.fillType = FillType.GOATMILK
	spec.info = {
		text = "",
		title = g_i18n:getText("fillType_goatmilk")
	}
	spec.milkingRobots = {}
	spec.husbandry = nil

	self.xmlFile:iterate("placeable.husbandry.goatmilk.milkingRobots.milkingRobot", function (_, key)
		local filename = Utils.getFilename(self.xmlFile:getValue(key .. "#filename", nil), self.baseDirectory)

		if filename == nil then
			Logging.xmlWarning(self.xmlFile, "Milkingrobot filename missing for '%s'", key)

			return
		end

		local className = self.xmlFile:getValue(key .. "#class", "")
		local class = ClassUtil.getClassObject(className)

		if class == nil then
			Logging.xmlWarning(self.xmlFile, "Milkingrobot class '%s' not defined for '%s'", className, key)

			return
		end

		local linkNode = self.xmlFile:getValue(key .. "#linkNode", nil, self.components, self.i3dMappings)

		if linkNode == nil then
			Logging.xmlWarning(self.xmlFile, "Milkingrobot linkNode not defined for '%s'", key)

			return
		end

		local loadingTask = self:createLoadingTask(self)
		local robot = class.new(self, self.baseDirectory)

		robot:load(linkNode, filename, self.onMilkingRobotLoaded, self, {
			loadingTask
		})
		table.insert(spec.milkingRobots, robot)
	end)
end

function PlaceableHusbandryGoatMilk:onMilkingRobotLoaded(robot, args)
	local task = unpack(args)

	self:finishLoadingTask(task)
end

function PlaceableHusbandryGoatMilk:onDelete()
	local spec = self.spec_husbandryGoatMilk

	if spec.milkingRobots ~= nil then
		for _, robot in ipairs(spec.milkingRobots) do
			robot:delete()
		end

		spec.milkingRobots = {}
	end
end

function PlaceableHusbandryGoatMilk:onFinalizePlacement()
	local spec = self.spec_husbandryGoatMilk

	if not self:getHusbandryIsFillTypeSupported(spec.fillType) then
		Logging.warning("Missing filltype 'goatmilk' in husbandry storage!")
	end

	for _, robot in ipairs(spec.milkingRobots) do
		robot:finalizePlacement()
	end
end

function PlaceableHusbandryGoatMilk:updateOutput(superFunc, foodFactor, productionFactor, globalProductionFactor)
	local spec = self.spec_husbandryGoatMilk

	if self.isServer and spec.litersPerHour > 0 then
		local liters = productionFactor * globalProductionFactor * spec.litersPerHour * g_currentMission.environment.timeAdjustment

		self:addHusbandryFillLevelFromTool(self:getOwnerFarmId(), liters, spec.fillType, nil, nil, nil)
	end

	superFunc(self, foodFactor, productionFactor, globalProductionFactor)
end

function PlaceableHusbandryGoatMilk:updateProduction(superFunc, foodFactor)
	local spec = self.spec_husbandryGoatMilk
	local factor = superFunc(self, foodFactor)
	local freeCapacity = self:getHusbandryFreeCapacity(spec.fillType)

	if freeCapacity <= 0 then
		factor = 0
	end

	return factor
end

function PlaceableHusbandryGoatMilk:onHusbandryAnimalsUpdate(clusters)
	local spec = self.spec_husbandryGoatMilk
	spec.litersPerHour = 0

	for _, cluster in ipairs(clusters) do
		local subType = g_currentMission.animalSystem:getSubTypeByIndex(cluster.subTypeIndex)

		if subType ~= nil then
			local milk = subType.output.milk

			if milk ~= nil then
				local age = cluster:getAge()
				local litersPerAnimals = milk:get(age)
				local litersPerDay = litersPerAnimals * cluster:getNumAnimals()
				spec.litersPerHour = spec.litersPerHour + litersPerDay / 24
			end
		end
	end
end

function PlaceableHusbandryGoatMilk:updateInfo(superFunc, infoTable)
	local spec = self.spec_husbandryGoatMilk

	superFunc(self, infoTable)

	local fillLevel = self:getHusbandryFillLevel(spec.fillType)
	spec.info.text = string.format("%d l", fillLevel)

	table.insert(infoTable, spec.info)
end

function PlaceableHusbandryGoatMilk:getConditionInfos(superFunc)
	local spec = self.spec_husbandryGoatMilk
	local infos = superFunc(self)
	local info = {}
	local fillType = g_fillTypeManager:getFillTypeByIndex(spec.fillType)
	info.title = fillType.title
	info.value = self:getHusbandryFillLevel(spec.fillType)
	local capacity = self:getHusbandryCapacity(spec.fillType)
	local ratio = 0

	if capacity > 0 then
		ratio = info.value / capacity
	end

	info.ratio = MathUtil.clamp(ratio, 0, 1)
	info.invertedBar = true

	table.insert(infos, info)

	return infos
end
